#include "MeshLoader.h"
#include <set>
#include <list>
#include <unordered_set>
#include <array>
std::vector<Node>& MeshLoader::GetNodes() {
	return nodes;
}

std::vector<FiniteElement>& MeshLoader::GetFE() {
	return finiteElements;
}

std::vector<BoundaryFiniteElement>& MeshLoader::GetBFE() {
	return boundaryFiniteElements;
}

void MeshLoader::Print_Data()
{
    std::cout << &GetNodes() << std::endl
        << &GetFE() << std::endl
        << &GetBFE() << std::endl;
}


std::vector<FiniteElement> MeshLoader::findFE(size_t ID1, size_t ID2, size_t ID3) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2, ID3](FiniteElement& cur_elem) {
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID3) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;

}


std::vector<FiniteElement> MeshLoader::findEdge(size_t ID1, size_t ID2) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2](FiniteElement& cur_elem) {
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;

}

std::vector<Node> MeshLoader::getSurfaceID(size_t s_id) {

        std::vector<Node> answer;
        auto current_iter = boundaryFiniteElements.begin();
        while (current_iter != boundaryFiniteElements.end()) {
            current_iter = std::find_if(current_iter, boundaryFiniteElements.end(), [s_id](const BoundaryFiniteElement& s) { return s.ID_surface == s_id; });
            if (current_iter != boundaryFiniteElements.end())
            {
                for (const auto& it : current_iter->node_boundary) {
                    auto node_by_ID = std::find_if(nodes.begin(), nodes.end(), [it](Node n) { return n.ID_node == it; });
                    answer.push_back(*node_by_ID);
                }
                ++current_iter;
            }
        }
        return answer;
    }


std::vector<Node> MeshLoader::getMaterialID(size_t ID) {
    std::list<Node> result;
    auto element_by_material = nodes.begin();
    auto last_element = nodes.end();
    while (element_by_material != last_element) {
        element_by_material = std::find_if(element_by_material, last_element, [ID](Node e) { return e.ID_node == ID; });
        if (element_by_material != last_element) {
            result.push_back(*element_by_material);
            ++element_by_material;
        }
    }
    return std::vector<Node>(result.begin(), result.end());
}

std::vector<std::set<size_t>>  MeshLoader::GetNodesNeighbors()
{
    std::vector<std::set<size_t>> result(nodes.size());
    for (const auto& it_FiniteElement : boundaryFiniteElements)
    {
        for (auto& it_nodes : it_FiniteElement.node_boundary)
        {
            result[it_nodes - 1].insert(it_FiniteElement.node_boundary.begin(), it_FiniteElement.node_boundary.end());
            result[it_nodes - 1].erase(it_nodes);
        }
    }

    return result;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteID(size_t ID) {
    std::list<BoundaryFiniteElement> result;
    auto surface_by_ID = boundaryFiniteElements.begin();
    auto last_element = boundaryFiniteElements.end();
    while (surface_by_ID != last_element) {
        surface_by_ID = std::find_if(surface_by_ID, last_element, [ID](BoundaryFiniteElement s) { return s.ID_surface == ID; });
        if (surface_by_ID != last_element) {
            result.push_back(*surface_by_ID);
            ++surface_by_ID;
        }
    }
    return std::vector<BoundaryFiniteElement>(result.begin(), result.end());
}

Node MeshLoader::Get_Node_by_ID(size_t n_id){
    for (const auto& it : nodes)
        if (it.ID_node== n_id)
            return it;
    return Node(0, std::vector<float>(), false);
}

void MeshLoader::InsertNewNodes(){
    std::unordered_set<Edge, EdgeHash> edge_container;
    for (auto& it : boundaryFiniteElements)
    {
        std::vector<int> current_iter(it.ID_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end() - 1; i++)
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge current_edge(*i, *j);
                auto search_edge = edge_container.find(current_edge);
                Node left_node = Get_Node_by_ID(*i);
                Node right_node = Get_Node_by_ID(*j);
                std::vector<float> x = { (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2 };
                Node center_node(nodes.size() + 1, x, false);

                if (search_edge == edge_container.end()) {
                    current_edge.self_id = nodes.size() + 1;
                    edge_container.insert(current_edge);
                    nodes.push_back(center_node);
                    it.node_boundary.emplace_back(current_edge.self_id);
                }
                else{
                    it.node_boundary.emplace_back(search_edge->self_id);
                }

            }
        }
    }

    for (auto& it : boundaryFiniteElements)
    {
        std::vector<size_t> current_iter(it.node_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end() - 1; i++)
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge search_edge(*i, *j);
                auto search_point = edge_container.find(search_edge);
                if (search_point != edge_container.end())
                {
                    it.node_boundary.emplace_back(search_point->self_id);
                }
            }
        }
    }


}
